import logging
import requests


log = logging.getLogger(__name__)


class ArticleNotFound(RuntimeError):
    """ Article query returned no results """


class Client(requests.Session):
    """ Mediawiki API client """

    def __init__(self, lang="en"):
        super(Client, self).__init__()
        self.base_url = 'https://' + lang + '.wikipedia.org/w/api.php'

    def fetch_page(self, title, method='GET'):
        """ Query for page by title """
        params = {
            'prop': 'revisions',
            'format': 'json',
            'action': 'query',
            'explaintext': '',
            'titles': _parse_title(title),
            'rvprop': 'content',
        }
        req = self.request(method, self.base_url, params=params)
        req.raise_for_status()
        pages = req.json()["query"]["pages"]
        # use key from first result in 'pages' array
        page_id = list(pages.keys())[0]
        if page_id == '-1':
            raise ArticleNotFound('no matching articles returned')

        return pages[page_id]

def _parse_title(s):
    # extract title from, potentially, a URL
    return s.split('/')[-1].split('#')[0].split('?')[0]



# Template readers
import logging

import pycountry

from wikitables.util import ustr
from wikitables.models import Field


log = logging.getLogger('wikitables')


def read_template(node, translate_fn):
    if node.name == 'refn':
        log.debug('omitting refn subtext from field')
        return []

    for read_function in _tmpl_readers:
        value = read_function(node, translate_fn)
        if value:
            return value
    return []


def _read_unknown_template(node, translate_fn):
    del translate_fn

    # for unknown templates, concatenate all arg values
    _, args = _read_template_params(node)
    concat = ' '.join([ustr(x) for x in args])
    return [concat]


def _read_change_template(node, translate_fn):
    del translate_fn

    if node.name != 'change':
        return None
    params, args = _read_template_params(node)
    args = [int(ustr(a)) for a in args]

    if params.get('invert') == 'on':
        change = ((args[0] / args[1]) - 1) * 100
    else:
        change = ((args[1] / args[0]) - 1) * 100

    return [Field(node, args[0]), Field(node, args[1]), Field(node, change)]


def _read_flag_template(node, translate_fn):
    # read flag shorthand templates
    sname = ustr(node.name)
    try:
        country = pycountry.countries.lookup(sname)
        return [translate_fn(country.name)]
    except LookupError:
        pass


def _read_template_params(node):
    kvs, args = {}, []
    for param in node.params:
        if '=' in param:
            parts = param.split('=')
            kvs[parts[0]] = '='.join(parts[1:])
        else:
            args.append(param)
    return kvs, args

_tmpl_readers = [
    _read_change_template,
    _read_flag_template,
    _read_unknown_template
]



import sys
import json
import logging


log = logging.getLogger('wikitables')


def ftag(*args):
    return lambda node: node.tag in args


def jprint(obj):
    if isinstance(obj, str):
        obj = json.loads(obj)
    print(json.dumps(obj, indent=2, sort_keys=False, cls=TableJSONEncoder))


def guess_type(value):
    """ attempt to convert string value into numeric type """
    num_value = value.replace(',', '') # remove comma from potential numbers

    try:
        return int(num_value)
    except ValueError:
        pass

    try:
        return float(num_value)
    except ValueError:
        pass

    return value


def ustr(value):
    if sys.version_info < (3, 0):
        #py2
        try:
            # pylint: disable=undefined-variable
            return unicode(value).encode('utf-8')
        except UnicodeDecodeError:
            return str(value)
    else:
        return str(value)


class TableJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if hasattr(o, '__json__'):
            return o.__json__()
        return json.JSONEncoder.default(self, o)



# pylint: disable=useless-object-inheritance

import json

from wikitables.util import TableJSONEncoder


class Field(object):
    """
    Field within a table row
    attributes:
     - raw(mwparserfromhell.nodes.Node) - Unparsed field Wikicode
     - value(str) - Parsed field value as string
    """
    def __init__(self, node, value, attrs=None):
        if attrs is None:
            attrs = {}
        self.raw = node
        self.value = value
        self.attrs = attrs

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)

    def __json__(self):
        return self.value


class Row(dict):
    """
    Single WikiTable row, mapping a field name(str) to wikitables.Field obj
    """
    def __init__(self, name, node):
        super(Row, self).__init__()
        self.name = name
        self.raw = node

    def json(self):
        return json.dumps(self, cls=TableJSONEncoder)

    @property
    def is_null(self):
        for _, item in self.items():
            if item.value != '':
                return False
        return True



# pylint: disable=invalid-name,useless-object-inheritance

import logging
from collections import defaultdict
import gettext

import pycountry
from mwparserfromhell.nodes.tag import Tag
from mwparserfromhell.nodes.template import Template
from mwparserfromhell.nodes.wikilink import Wikilink

from wikitables.models import Field, Row
from wikitables.util import ftag, ustr, guess_type
from wikitables.templates import read_template


log = logging.getLogger('wikitables')

ignore_attrs = ['group="Note"']


class FieldReader(object):
    """ Stateful Field value reader """

    def __init__(self, lang='en'):
        self.lang = lang
        try:
            language_translation = gettext.translation(
                'iso3166', pycountry.LOCALES_DIR, languages=[lang]
            )
            language_translation.install()
        except FileNotFoundError:
            language_translation = gettext
        self.translate_fn = language_translation.gettext

        self._attrs = {} # node attribute state

    def parse(self, node):
        """
        Return generator yielding Field objects for a given node
        """
        self._attrs = {}
        vals = []
        yielded = False

        for x in self._read_parts(node):
            if isinstance(x, Field):
                yielded = True
                x.attrs = self._attrs
                yield x
            else:
                vals.append(ustr(x).strip(' \n\t'))

        joined = ' '.join([x for x in vals if x])
        if joined:
            yielded = True
            yield Field(node, guess_type(joined), self._attrs)

        if not yielded:
            yield Field(node, "", self._attrs)

    def _read_parts(self, n):
        for a in getattr(n, 'attributes', []):
            self._attrs[ustr(a.name)] = ustr(a.value)

        if hasattr(n, 'contents') and hasattr(n.contents, 'nodes'):
            for subnode in n.contents.nodes:
                for x in self._read_parts(subnode):
                    yield x
        else:
            for x in self._read_part(n):
                yield x

    def _read_part(self, node):
        if isinstance(node, Template):
            for x in read_template(node, self.translate_fn):
                yield x
            return
        if isinstance(node, Tag):
            if not self._exclude_tag(node):
                yield node.contents.strip_code()
            return
        if isinstance(node, Wikilink):
            if node.text:
                yield node.text
            else:
                yield node.title
            return

        yield node

    @staticmethod
    def _exclude_tag(node):
        # exclude tag nodes with attributes in ignore_attrs
        n_attrs = [x.strip() for x in node.attributes]
        for a in n_attrs:
            if a in ignore_attrs:
                return True

        # exclude tag nodes without contents
        if not node.contents:
            return True

        return False


class RowReader(object):
    """ Stateful Row reader """

    def __init__(self, tname, head, lang='en'):
        self.head = head
        self.lang = lang
        self._idx = 0
        self._tname = tname
        # track spanned fields across rows
        self._span = {}
        self._nspan = defaultdict(int)
        self._freader = FieldReader(lang)

    def parse(self, *nodes):
        """
        Parse one or more `tr` nodes, yielding wikitables.Row objects
        """
        for n in nodes:
            if not n.contents:
                continue
            row = self._parse(n)
            if not row.is_null:
                yield row

    def _parse(self, node):
        rname = '%s[%s]' % (self._tname, self._idx)
        self._idx += 1
        r = Row(rname, node)
        cols = node.contents.ifilter_tags(matches=ftag('th', 'td'))
        fields = [f for col in cols for f in self._freader.parse(col)]

        for col_name in self.head:
            if self._nspan[col_name]:
                r[col_name] = self._span[col_name]
                self._nspan[col_name] -= 1
                continue

            if not fields:
                log.warning('%s: missing field for column [%s]', r.name, col_name)
                continue

            f = fields.pop(0)
            if 'rowspan' in f.attrs:
                self._span[col_name] = f
                self._nspan[col_name] = int(f.attrs['rowspan'])-1

            r[col_name] = f

        for f in fields:
            log.warning('%s: dropping field from unknown column: %s', r.name, f)

        return r



import json
import logging

import mwparserfromhell as mwp

from wikitables.client import Client
from wikitables.readers import RowReader
from wikitables.util import TableJSONEncoder, ftag, ustr


log = logging.getLogger('wikitables')


def import_tables(article, lang='en'):
    client = Client(lang)
    page = client.fetch_page(article)
    body = page['revisions'][0]['*']

    ## parse for tables
    raw_tables = mwp.parse(body).filter_tags(matches=ftag('table'))

    def _table_gen():
        for idx, table in enumerate(raw_tables):
            name = '%s[%s]' % (page['title'], idx)
            yield WikiTable(name, table, lang)

    return list(_table_gen())


class WikiTable():
    """
    Parsed Wikipedia table
    attributes:
     - name(str): Table name in the format <article_name>[<table_index>]
     - head(list): List of parsed column names as strings
     - rows(list): List of <wikitables.Row> objects
    """
    def __init__(self, name, raw_table, lang='en'):
        self.name = ustr(name)
        self.lang = lang
        self.rows = []
        self._head = []
        self._node = raw_table
        self._tr_nodes = raw_table.contents.filter_tags(matches=ftag('tr'))
        self._read_header()
        self._read_rows()

    def json(self):
        return json.dumps(self.rows, cls=TableJSONEncoder)

    @property
    def head(self):
        return self._head

    @head.setter
    def head(self, val):
        if not isinstance(val, list):
            raise ValueError('table head must be provided as list')
        self._head = val
        self._read_rows()

    def __repr__(self):
        return "<WikiTable '%s'>" % self.name

    def _log(self, value):
        log.debug('%s: %s', self.name, value)

    def _read_rows(self):
        reader = RowReader(self.name, self._head, self.lang)
        self.rows = list(reader.parse(*self._tr_nodes))
        self._log('parsed %d rows %d cols' % (len(self.rows), len(self._head)))

    def _read_header(self):
        # read header
        header_nodes = self._find_header_flat()
        if not header_nodes:
            header_nodes = self._find_header_row()
        if not header_nodes:
            header_nodes = self._make_default_header()

        for header_node in header_nodes:
            field_name = header_node.contents.strip_code().strip(' ')
            self._head.append(ustr(field_name))

    def _find_header_flat(self):
        """
        Find header elements in a table, if possible. This case handles
        situations where '<th>' elements are not within a row('<tr>')
        """
        nodes = self._node.contents.filter_tags(matches=ftag('th'), recursive=False)
        if not nodes:
            return None
        self._log('found header outside rows (%d <th> elements)' % len(nodes))
        return nodes

    def _find_header_row(self):
        """
        Evaluate all rows and determine header position, based on
        greatest number of 'th' tagged elements
        """
        th_max = 0
        header_idx = 0
        for idx, tr_node in enumerate(self._tr_nodes):
            th_count = len(tr_node.contents.filter_tags(matches=ftag('th')))
            if th_count > th_max:
                th_max = th_count
                header_idx = idx

        if not th_max:
            return None

        self._log('found header at row %d (%d <th> elements)' % \
                    (header_idx, th_max))

        header_row = self._tr_nodes.pop(header_idx)
        return header_row.contents.filter_tags(matches=ftag('th'))

    def _make_default_header(self):
        """
        Return a generic placeholder header based on the tables column count
        """
        td_max = 0

        for tr_node in self._tr_nodes:
            td_count = len(tr_node.contents.filter_tags(matches=ftag('td')))
            if td_count > td_max:
                td_max = td_count

        self._log('creating default header (%d columns)' % td_max)
        return ['column%d' % n for n in range(0, td_max)]



from setuptools import setup

exec(open('wikitables/version.py').read())

setup(name='wikitables',
      version=version,
      packages=['wikitables'],
      description='Import tables from any Wikipedia article',
      author='Bradley Cicenas',
      author_email='bradley@vektor.nyc',
      url='https://github.com/bcicen/wikitables',
      install_requires=[
          'mwparserfromhell>=0.4.3',
          'requests>=2.9.1',
          'pycountry>=20.7.3'
      ],
      license='http://opensource.org/licenses/MIT',
      classifiers=(
          'Natural Language :: English',
          'Programming Language :: Python',
          'Intended Audience :: Developers',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: MIT License',
      ),
      keywords='wikipedia data cli commandline',
      entry_points={'console_scripts': ['wikitables = wikitables.cli:main']}
)



import sys
import json
import logging
from argparse import ArgumentParser

from wikitables.version import version
from wikitables.util import TableJSONEncoder
from wikitables import import_tables
from wikitables.util import jprint


log = logging.getLogger('wikitables')


def main():
    parser = ArgumentParser(description='wikitables v%s' % version)
    parser.add_argument('-l', '--lang',
                        dest='lang',
                        help='article language (default: %(default)s)',
                        default='en')
    parser.add_argument('-p', '--pretty',
                        action='store_true',
                        help='pretty-print json output')
    parser.add_argument('-d', '--debug',
                        action='store_true',
                        help='enable debug output')
    parser.add_argument('article', help='article title')

    args = parser.parse_args()

    if not args.article:
        print('usage: wikitables <article title>')
        sys.exit(1)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
        log.debug('debug logging enabled')
    else:
        logging.basicConfig(level=logging.WARN)

    tables = import_tables(args.article, lang=args.lang)
    tables_dict = {table.name: table.rows for table in tables}
    if args.pretty:
        jprint(tables_dict)
    else:
        print(json.dumps(tables_dict, cls=TableJSONEncoder))
